/*
 *   jack-ladspa-host
 *    
 *   Copyright (C) Robert Ham 2002 (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>

#include "chain.h"
#include "gui.h"
#include "xmalloc.h"
#include "plugin.h"

#define CONTROL_FIFO_SIZE   256


/** connect up the ladspa instance's input buffers to the previous
    plugin's audio memory.  make sure to check that plugin->prev
    exists. */
void connect_input_ports (ladspa_chain_link_t * plugin) {
  if (!plugin || !plugin->prev) return;
  if (plugin->mono) {
    plugin->descriptor->
      connect_port (plugin->holder.mono_pair.left->instance,
                    plugin->port_indicies[IN_L],
                    plugin->prev->audio_memory[0]);
    plugin->descriptor->
      connect_port (plugin->holder.mono_pair.right->instance,
                    plugin->port_indicies[IN_R],
                    plugin->prev->audio_memory[1]);
  } else {
    plugin->descriptor->
      connect_port (plugin->holder.stereo->instance,
                    plugin->port_indicies[IN_L],
                    plugin->prev->audio_memory[0]);
    plugin->descriptor->
      connect_port (plugin->holder.stereo->instance,
                    plugin->port_indicies[IN_R],
                    plugin->prev->audio_memory[1]);
  }
}

/** connect up a plugin's output ports to its own audio_memory output memory */
void connect_output_ports (ladspa_chain_link_t * plugin) {
  if (!plugin) return;
  if (plugin->mono) {
    plugin->descriptor->
      connect_port (plugin->holder.mono_pair.left->instance,
                    plugin->port_indicies[OUT_L],
                    plugin->audio_memory[0]);
    plugin->descriptor->
      connect_port (plugin->holder.mono_pair.right->instance,
                    plugin->port_indicies[OUT_R],
                    plugin->audio_memory[1]);
  } else {
    plugin->descriptor->
      connect_port (plugin->holder.stereo->instance,
                    plugin->port_indicies[OUT_L],
                    plugin->audio_memory[0]);
    plugin->descriptor->
      connect_port (plugin->holder.stereo->instance,
                    plugin->port_indicies[OUT_R],
                    plugin->audio_memory[1]);
  }
}

void add_plugin (process_info_t * procinfo, ladspa_chain_link_t * plugin) {

  /* sort out list pointers */
  plugin->next = NULL;
  plugin->prev = procinfo->chain_end;
  
  if (procinfo->chain_end) {
    procinfo->chain_end->next = plugin;
  } else {
    procinfo->chain = plugin;
  }
  
  procinfo->chain_end = plugin;
  
  /* previous was moved off the bottom; reconnect the audio memory */
/*  if (plugin->prev) {
    connect_input_ports (plugin);
  }*/
}

/** find a plugin given its index in the chain */
ladspa_chain_link_t * get_plugin (process_info_t * procinfo,
                                  gint plugin_index) {
  gint i = 0;
  ladspa_chain_link_t * plugin = procinfo->chain;
  
  while (i < plugin_index) {
    plugin = plugin->next;
    i++;
  }
  
  return plugin;
}

/** remove a plugin from the chain */
ladspa_chain_link_t * remove_plugin (process_info_t * procinfo, gint plugin_index) {
  ladspa_chain_link_t * plugin;
  
  /* find the plugin in the chain */
  plugin = get_plugin (procinfo, plugin_index);
  
  /* sort out chain pointers */
  if (plugin->prev) {
    plugin->prev->next = plugin->next;
  } else {
    procinfo->chain = plugin->next;
  }
  
  if (plugin->next) {
    plugin->next->prev = plugin->prev;
/*    if (plugin->prev) {
      connect_input_ports (plugin->next);
    }*/
  } else {
    procinfo->chain_end = plugin->prev;
  }

  return plugin;
}

/** enable/disable a plugin */
void ablise_plugin (
  process_info_t * procinfo,
  gint plugin_index,
  gint enable) {
  
  ladspa_chain_link_t * plugin;
  
  plugin = get_plugin (procinfo, plugin_index);
  
  plugin->enabled = enable;
  
  /* sort out the audio buffers */
/*  if (plugin->prev) {
    connect_output_ports (plugin->prev);
  }
  connect_input_ports (plugin);
  connect_output_ports (plugin);
  if (plugin->next) {
    connect_input_ports (plugin->next);
  }*/
  
}

/** move a plugin up or down one place in the chain */
void move_plugin (
  process_info_t * procinfo,
  gint plugin_index,
  gint up) {
  
  /* specified plugin */
  ladspa_chain_link_t * plugin;
  /* other plugins in the chain */
  ladspa_chain_link_t * pp = NULL, * p, * n, * nn = NULL;
  
  plugin = get_plugin (procinfo, plugin_index);
  
  /* note that we should never recieve an illogical move request
     ie, there will always be at least 1 plugin before for an up
     request or 1 plugin after for a down request */
  
  /* these are pointers to the plugins surrounding the specified one:
      { pp, p, plugin, n, nn } which makes things much clearer than
      tptr, tptr2 etc */
  p = plugin->prev;
  if (p) pp = p->prev;
  n = plugin->next;
  if (n) nn = n->next;
  
  if (up) {
    if (!p) return;
  
    if (pp) {
      pp->next = plugin;
    } else {
      procinfo->chain = plugin;
    }
    
    p->next = n;
    p->prev = plugin;
    
    plugin->prev = pp;
    plugin->next = p;
    
    if (n) {
      n->prev = p;
    } else {
      procinfo->chain_end = p;
      
      /* we've been moved off the bottom, which means the output buffers
         would have been pointing at jack buffers.  we need to reconnect
         them to the audio_memory */
/*      connect_output_ports (plugin);*/
    }
    
/*    if (pp) connect_input_ports (plugin);
    connect_input_ports (p);
    if (n) connect_input_ports (n);*/
    
  } else {
    if (!n) return;
  
    if (p) {
      p->next = n;
    } else {
      procinfo->chain = n;
    }
    
    n->prev = p;
    n->next = plugin;
    
    plugin->prev = n;
    plugin->next = nn;

    if (nn) {
      nn->prev = plugin;
    } else {
      procinfo->chain_end = plugin;
      
      /* n has been moved off of the bottom; reconnect its output buffers */
/*      connect_output_ports (n); */
    }
    
/*    if (p) connect_input_ports (n);
    connect_input_ports (plugin);
    if (nn) connect_input_ports (nn);*/
    
  }
}

/** exchange an existing plugin for a newly created one */
ladspa_chain_link_t * change_plugin (
  process_info_t * procinfo,
  gint plugin_index,
  ladspa_chain_link_t * new_plugin) {
  
  ladspa_chain_link_t * plugin;
  
  plugin = get_plugin (procinfo, plugin_index);
  
  new_plugin->next = plugin->next;
  new_plugin->prev = plugin->prev;
  
  if (plugin->prev) {
    plugin->prev->next = new_plugin;
/*    connect_input_ports (new_plugin);*/
  } else {
    procinfo->chain = new_plugin;
  }
  
  if (plugin->next) {
    plugin->next->prev = new_plugin;
/*    connect_input_ports (plugin->next);*/
  } else {
    procinfo->chain_end = new_plugin;
  }
  
  return plugin;
}


/******************************************
 ************* non RT stuff ***************
 ******************************************/


int init_instance (ladspa_holder_t * holder,
                    plugin_t * plugin) {
  unsigned long i;
  
  /* insantiate the plugin.. woo! */
  holder->instance = plugin->descriptor->
    instantiate (plugin->descriptor, sample_rate);
    
  if (!holder->instance) {
    fprintf(stderr, "%s: could not initialise plugin '%s' from object file '%s'\n",
            __FUNCTION__, plugin->name, plugin->object_file->filename);
    return 1;
  }
  
  if (plugin->control_port_count > 0) {
    holder->control_fifos =
      xmalloc (sizeof (lff_t) * plugin->control_port_count);
    holder->control_memory =
      xmalloc (sizeof (LADSPA_Data) * plugin->control_port_count);
  
    for (i = 0; i < plugin->control_port_count; i++) {
    
      /* init the fifo */
      lff_init (holder->control_fifos + i, CONTROL_FIFO_SIZE, sizeof (LADSPA_Data));
    
      /* connect up the control memory */
/*      printf("%s: setting control port %ld of plugin '%s' to %p; plugin has %ld ports in file '%s'\n",
             __FUNCTION__, plugin->control_port_indicies[i],
             plugin->name, (void *) (holder->control_memory + i),
             plugin->descriptor->PortCount, plugin->object_file->filename);*/
      plugin->descriptor->
        connect_port (holder->instance, plugin->control_port_indicies[i],
                      holder->control_memory + i);
                      
      /* set it to the default for now */
      holder->control_memory[i] =
        global_plugin_collection.default_plugin_settings[plugin->collection_index][i];
    }
  } else {
    holder->control_fifos = NULL;
    holder->control_memory = NULL;
  }

  /* activate it */
  if (plugin->descriptor->activate) {
    plugin->descriptor->activate (holder->instance);
  }
  
  return 0;
}

LADSPA_Data unused_control_port_output;

/** initialise a new chain link */
int init_chain_link (ladspa_chain_link_t * chain_link,
                      plugin_t * plugin) {
  unsigned long i;
  int err;
  
  chain_link->plugin = plugin;
  chain_link->mono = plugin->mono;
  chain_link->enabled = FALSE;
  chain_link->descriptor = plugin->descriptor;
  chain_link->port_indicies = plugin->port_indicies;
  chain_link->control_port_count = plugin->control_port_count;
  chain_link->control_port_indicies = plugin->control_port_indicies;
  chain_link->next = NULL;
  chain_link->prev = NULL;
  
  chain_link->audio_memory[0] = xmalloc (sizeof (LADSPA_Data) * buffer_size);
  chain_link->audio_memory[1] = xmalloc (sizeof (LADSPA_Data) * buffer_size);
  
  if (plugin->mono) {
  
    chain_link->holder.mono_pair.left = xmalloc (sizeof (ladspa_holder_t));
    err = init_instance (chain_link->holder.mono_pair.left, plugin);
    if (err) {
      free (chain_link->holder.mono_pair.left);
      free (chain_link->audio_memory[0]);
      free (chain_link->audio_memory[1]);
      return err;
    }
    
    
    chain_link->holder.mono_pair.right = xmalloc (sizeof (ladspa_holder_t));
    err = init_instance (chain_link->holder.mono_pair.right, plugin);
    if (err) {
      if (chain_link->holder.mono_pair.left->control_fifos)
        free (chain_link->holder.mono_pair.left->control_fifos);
      if (chain_link->holder.mono_pair.left->control_memory)
        free (chain_link->holder.mono_pair.left->control_fifos);
      free (chain_link->holder.mono_pair.right);
      free (chain_link->holder.mono_pair.left);
      free (chain_link->audio_memory[0]);
      free (chain_link->audio_memory[1]);
      return err;
    }
    
    /* connect up the audio memory */
    
    
    plugin->descriptor->connect_port (chain_link->holder.mono_pair.left->instance,
                                      plugin->port_indicies[OUT_L],
                                      chain_link->audio_memory[0]);
    plugin->descriptor->connect_port (chain_link->holder.mono_pair.right->instance,
                                      plugin->port_indicies[OUT_R],
                                      chain_link->audio_memory[1]);
  } else {
    chain_link->holder.stereo = xmalloc (sizeof (ladspa_holder_t));
    err = init_instance (chain_link->holder.stereo, plugin);
    if (err) {
      free (chain_link->holder.stereo);
      free (chain_link->audio_memory[0]);
      free (chain_link->audio_memory[1]);
      return err;
    }

    plugin->descriptor->connect_port (chain_link->holder.stereo->instance,
                                      plugin->port_indicies[OUT_L],
                                      chain_link->audio_memory[0]);
    plugin->descriptor->connect_port (chain_link->holder.stereo->instance,
                                      plugin->port_indicies[OUT_R],
                                      chain_link->audio_memory[1]);
  }
  
  /* connect up the control output ports (which we have to do) */
  for (i = 0; i < plugin->descriptor->PortCount; i++) {
    if (LADSPA_IS_PORT_CONTROL(plugin->descriptor->PortDescriptors[i]) &&
        LADSPA_IS_PORT_OUTPUT (plugin->descriptor->PortDescriptors[i])) {
      printf ("%s: unused control output port at index %ld on plugin '%s'\n",
              __FUNCTION__, i, plugin->name);
      if (plugin->mono) {
        plugin->descriptor->connect_port (chain_link->holder.mono_pair.left,
                                          i, &unused_control_port_output);
        plugin->descriptor->connect_port (chain_link->holder.mono_pair.right,
                                          i, &unused_control_port_output);
      } else {
        plugin->descriptor->connect_port (chain_link->holder.stereo,
                                          i, &unused_control_port_output);
      }
    }
  }
  
  return 0;
}

void destroy_chain_link (ladspa_chain_link_t * chain_link) {
/*  unsigned long i;*/
  /* deactivate and unload the plugins and free up their holders' memory */
  if (chain_link->mono) {
    if (chain_link->descriptor->deactivate) {
      chain_link->descriptor->deactivate (chain_link->holder.mono_pair.left->instance);
      chain_link->descriptor->deactivate (chain_link->holder.mono_pair.right->instance);
    }
    chain_link->descriptor->cleanup (chain_link->holder.mono_pair.left->instance);
    chain_link->descriptor->cleanup (chain_link->holder.mono_pair.right->instance);
    
/*    FIXME: wtf is going on here?  these segfault
    for (i = 0; i < chain_link->descriptor->PortCount; i++) {
       lff_free (chain_link->holder.mono_pair.left->control_fifos + i); 
       lff_free (chain_link->holder.mono_pair.right->control_fifos + i);
    }*/
    free (chain_link->holder.mono_pair.left->control_fifos);
    free (chain_link->holder.mono_pair.right->control_fifos);
    free (chain_link->holder.mono_pair.left->control_memory);
    free (chain_link->holder.mono_pair.right->control_memory);
    free (chain_link->holder.mono_pair.left);
    free (chain_link->holder.mono_pair.right);
  } else {
    if (chain_link->descriptor->deactivate) {
      chain_link->descriptor->deactivate (chain_link->holder.stereo->instance);
    }
    chain_link->descriptor->cleanup (chain_link->holder.stereo->instance);
    
/*    for (i = 0; i < chain_link->descriptor->PortCount; i++) {
      lff_free (chain_link->holder.stereo->control_fifos + i);
    }*/
    free (chain_link->holder.stereo->control_fifos);
    free (chain_link->holder.stereo->control_memory);
    free (chain_link->holder.stereo);
  }
  
  free (chain_link->audio_memory[0]);
  free (chain_link->audio_memory[1]);
  
  free (chain_link);
}

